package com.yexuejc.base.util;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.RandomAccessFile;
import java.math.BigInteger;
import java.nio.MappedByteBuffer;
import java.nio.channels.FileChannel;
import java.nio.charset.Charset;
import java.nio.file.Files;
import java.nio.file.Path;
import java.security.MessageDigest;
import java.security.NoSuchAlgorithmException;
import java.util.ArrayList;
import java.util.Base64;
import java.util.List;
import java.util.logging.Logger;
import java.util.zip.CRC32;

import com.fasterxml.jackson.databind.MappingIterator;
import com.fasterxml.jackson.dataformat.csv.CsvMapper;
import com.fasterxml.jackson.dataformat.csv.CsvSchema;
import com.yexuejc.base.annotation.CsvToBean;
import com.yexuejc.base.pojo.ReadFileBean;
import io.jsonwebtoken.lang.Assert;

/**
 * 文件工具类
 *
 * @author maxf:yexue
 * @className FileUtil
 * @description 工具类
 * @time 2017年11月3日 下午3:12:49
 */
public class FileUtil {
    static Logger logger = Logger.getLogger(FileUtil.class.getName());

    public FileUtil() {
    }

    private static final String TYPE_TAR_GZ = ".tar.gz";
    private static final String TYPE_CSV = ".csv";
    private static final String TAR_GZ = "tar.gz";

    /**
     * 获取文件类型：不适合所有
     * <p>
     * 根据文件名称截取.后的文件格式
     * </p>
     *
     * @param fileName
     * @return
     */
    public static String getFileType(String fileName) {
        try {
            if (fileName.lastIndexOf(TYPE_TAR_GZ) > 0) {
                return TAR_GZ;
            }
            return fileName.substring(fileName.lastIndexOf(".") + 1);
        } catch (Exception e) {
            logger.severe("file doesn't exist or is not a file");
        }
        return null;
    }

    /**
     * 判断文件是否存在,不存在就创建一个空的
     *
     * @param file
     */
    public static void judeFileExists(File file) {

        if (file.exists()) {
            logger.severe("file exists");
        } else {
            logger.info("file not exists, create it ...");
            try {
                file.createNewFile();
            } catch (IOException e) {
                logger.severe("file create fail");
                e.printStackTrace();
            }
        }

    }

    /**
     * <pre>
     * 判断文件夹是否存在, 不存在创建【采用mkdirs】
     * 相关解释：
     * 1、File类的createNewFile根据抽象路径创建一个新的空文件，当抽象路径制定的文件存在时，创建失败
     * 2、File类的mkdir方法根据抽象路径创建目录
     * 3、File类的mkdirs方法根据抽象路径创建目录，包括创建必需但不存在的父目录
     * 4、File类的createTempFile方法创建临时文件，可以制定临时文件的文件名前缀、后缀及文件所在的目录，如果不指定目录，则存放在系统的临时文件夹下。
     * 5、除mkdirs方法外，以上方法在创建文件和目录时，必须保证目标文件不存在，而且父目录存在，否则会创建失败
     * </pre>
     *
     * @return 创建成功、失败
     */
    public static boolean judeDirExists(File file) {
        if (file.exists()) {
            if (file.isDirectory()) {
                logger.severe("dir exists");
            } else {
                logger.severe("the same name file exists, can not create dir");
            }
        } else {
            logger.info("dir not exists, create it ...");
            return file.mkdirs();
        }
        return false;
    }

    /**
     * 获取文件sha1
     *
     * @param file
     * @return
     */
    public static String sha1(File file) {
        FileInputStream in = null;
        try {
            in = new FileInputStream(file);
            MessageDigest digest = MessageDigest.getInstance("SHA-1");
            byte[] buffer = new byte[1024 * 1024 * 10];

            int len = 0;
            while ((len = in.read(buffer)) > 0) {
                digest.update(buffer, 0, len);
            }
            String sha1 = new BigInteger(1, digest.digest()).toString(16);
            int length = 40 - sha1.length();
            if (length > 0) {
                for (int i = 0; i < length; i++) {
                    sha1 = "0" + sha1;
                }
            }
            return sha1;
        } catch (NoSuchAlgorithmException e) {
            logger.severe("system algorithm error.");
            e.printStackTrace();
        } catch (FileNotFoundException e) {
            logger.severe("file doesn't exist or is not a file");
            e.printStackTrace();
        } catch (IOException e) {
            logger.severe("The operation file is an IO exception.");
            e.printStackTrace();
        } finally {
            try {
                if (in != null) {
                    in.close();
                }
            } catch (IOException e) {
                logger.severe("close FileInputStream IO exception.");
            }
        }
        return null;
    }

    /***
     * 计算SHA1码
     *
     * @return String 适用于上G大的文件
     * @throws NoSuchAlgorithmException
     * */
    public static String sha1ByBigFile(File file) {
        MessageDigest messagedigest = null;
        try {
            messagedigest = MessageDigest.getInstance("SHA-1");
            FileInputStream in = new FileInputStream(file);
            FileChannel ch = in.getChannel();
            MappedByteBuffer byteBuffer = ch.map(FileChannel.MapMode.READ_ONLY, 0, file.length());
            messagedigest.update(byteBuffer);
            return StrUtil.toHex(messagedigest.digest());
        } catch (NoSuchAlgorithmException e) {
            logger.severe("system algorithm error.");
            e.printStackTrace();
        } catch (FileNotFoundException e) {
            logger.severe("file doesn't exist or is not a file");
            e.printStackTrace();
        } catch (IOException e) {
            logger.severe("The operation file is an IO exception.");
            e.printStackTrace();
        }
        return null;
    }

    /**
     * 文件md5
     *
     * @param file
     * @return
     */
    public static String md5(File file) {
        FileInputStream in = null;
        try {
            in = new FileInputStream(file);
            MessageDigest digest = MessageDigest.getInstance("MD5");
            byte[] buffer = new byte[1024 * 1024 * 10];

            int len = 0;
            while ((len = in.read(buffer)) > 0) {
                digest.update(buffer, 0, len);
            }
            String md5 = new BigInteger(1, digest.digest()).toString(16);
            int length = 32 - md5.length();
            if (length > 0) {
                for (int i = 0; i < length; i++) {
                    md5 = "0" + md5;
                }
            }
            return md5;
        } catch (IOException e) {
            logger.severe("The operation file is an IO exception.");
        } catch (NoSuchAlgorithmException e) {
            logger.severe("system algorithm error.");
        } finally {
            try {
                if (in != null) {
                    in.close();
                }
            } catch (IOException e) {
                logger.severe("close FileInputStream IO exception.");
            }
        }
        return null;
    }

    /**
     * 对一个文件获取md5值
     *
     * @return md5串
     * @throws NoSuchAlgorithmException
     */
    public static String md5ByBigFile(File file) {

        MessageDigest messagedigest = null;
        try {
            messagedigest = MessageDigest.getInstance("MD5");
            FileInputStream in = new FileInputStream(file);
            FileChannel ch = in.getChannel();
            MappedByteBuffer byteBuffer = ch.map(FileChannel.MapMode.READ_ONLY, 0,
                    file.length());
            messagedigest.update(byteBuffer);
            return StrUtil.toHex(messagedigest.digest());
        } catch (NoSuchAlgorithmException e) {
            logger.severe("system algorithm error.");
            e.printStackTrace();
        } catch (FileNotFoundException e) {
            logger.severe("file doesn't exist or is not a file");
            e.printStackTrace();
        } catch (IOException e) {
            logger.severe("The operation file is an IO exception.");
            e.printStackTrace();
        }
        return null;
    }

    /**
     * 获取文件CRC32码
     *
     * @return String
     */
    public static String crc32(File file) {
        CRC32 crc32 = new CRC32();
        // MessageDigest.get
        FileInputStream fileInputStream = null;
        try {
            fileInputStream = new FileInputStream(file);
            byte[] buffer = new byte[8192];
            int length;
            while ((length = fileInputStream.read(buffer)) != -1) {
                crc32.update(buffer, 0, length);
            }
            return crc32.getValue() + "";
        } catch (FileNotFoundException e) {
            logger.severe("file doesn't exist or is not a file");
            e.printStackTrace();
            return null;
        } catch (IOException e) {
            logger.severe("The operation file is an IO exception.");
            e.printStackTrace();
            return null;
        } finally {
            try {
                if (fileInputStream != null) {
                    fileInputStream.close();
                }
            } catch (IOException e) {
                logger.severe("close FileInputStream IO exception.");
                e.printStackTrace();
            }
        }
    }

    /**
     * 获取文件base64
     *
     * @param file
     * @return
     */
    public static String base64ToStr(File file) {
        try {
            byte[] bytes = Files.readAllBytes(Path.of(file.getPath()));
            return Base64.getEncoder().encodeToString(bytes);
        } catch (IOException e) {
            logger.severe("The operation file is an IO exception.");
        }
        return null;
    }

    /**
     * base64转文件
     * <p>
     * <i>
     * 文件转base64请使用 {@link FileUtil#base64ToStr(File)}
     * </i>
     *
     * @param decode   {@link FileUtil#base64ToStr(File)} 的结果
     * @param fileName 保存文件名称（包含路径）
     * @return 返回保存地址
     */
    public static String base64ToFile(String decode, String fileName) {
        return base64ToFile(Base64.getDecoder().decode(decode.getBytes()), fileName);
    }

    /**
     * base64转文件
     * <p>
     * <i>
     * 文件转base64请使用 {@link FileUtil#base64ToStr(File)}
     * </i>
     *
     * @param decode   baseByte
     * @param fileName 文件名称（包含路径）
     * @return 返回保存地址
     */
    public static String base64ToFile(byte[] decode, String fileName) {

        FileOutputStream out = null;
        try {
            out = new FileOutputStream(fileName);
            out.write(decode);
        } catch (IOException ioe) {
            ioe.printStackTrace();
        } finally {
            try {
                if (out != null) {
                    out.close();
                }
            } catch (IOException e) {
                e.printStackTrace();
            }
        }
        return fileName;
    }

    /**
     * 获取文件大小 ：直接返回大小
     *
     * @param f
     * @return f.length()
     */
    public static long size(File f) {
        if (f.exists() && f.isFile()) {
            return f.length();
        } else {
            logger.info("file doesn't exist or is not a file");
        }
        return 0;
    }

    /**
     * 获取文件大小 : 用流的方式获取
     *
     * @param f
     * @return
     */
    public static long size4Stream(File f) {
        FileChannel fc = null;
        try {
            if (f.exists() && f.isFile()) {
                FileInputStream fis = new FileInputStream(f);
                fc = fis.getChannel();
                return fc.size();
            } else {
                logger.info("file doesn't exist or is not a file");
            }
        } catch (FileNotFoundException e) {
            logger.severe("file doesn't exist or is not a file");
        } catch (IOException e) {
            logger.severe("The operation file is an IO exception.");
        } finally {
            if (null != fc) {
                try {
                    fc.close();
                } catch (IOException e) {
                    logger.severe("close FileInputStream IO exception.");
                }
            }
        }
        return 0;
    }

    /**
     * 字符串（csv格式）转 对象
     *
     * @param data      转换的字符串 如
     *                  <p> ------------ </p>
     *                  <p> id,name,age </p>
     *                  <p> 1,zhangsan,18   </p>
     *                  <p> 2,lisi,20 </p>
     *                  <p> ------------ </p>
     * @param cls       需要转换的对象,含有<b>id,name,age</b>字段
     * @param delimiter 分隔符
     * @param <I>
     * @return
     */
    public static <I> List<I> readCsv(String data, Class<I> cls, char delimiter) throws IOException {
        CsvMapper csvMapper = new CsvMapper();
        CsvSchema csvSchema = CsvSchema.emptySchema().withHeader().withColumnSeparator(delimiter);
        MappingIterator<I> orderLines = csvMapper.readerFor(cls).with(csvSchema).readValues(data);
        return orderLines.readAll();
    }

    /**
     * 读取csv文件
     *
     * @param csvFilePath 文件地址
     * @param cls         读取转化的对象
     * @param hasHeader   是否存在header
     * @param delimiter   分隔符.默认【,】
     * @param <I>
     * @return
     */
    public static <I> List<I> readCsv(final String csvFilePath, Class<I> cls, boolean hasHeader, char delimiter) {
        if (!isFileExist(csvFilePath)) {
            throw new RuntimeException(String.format("解析用的csv：\u0020[%s] 文件不存在。", csvFilePath));
        }
        if (StrUtil.isEmpty(delimiter)) {
            delimiter = ',';
        }
        try {
            File csvFile = new File(csvFilePath);
            CsvMapper csvMapper = new CsvMapper();
            CsvSchema csvSchema = csvMapper.typedSchemaFor(cls).withStrictHeaders(hasHeader).withColumnSeparator(delimiter).withComments();
            MappingIterator<I> recordIterator = csvMapper.readerWithTypedSchemaFor(cls).with(csvSchema).readValues(csvFile);
            return recordIterator.readAll();
        } catch (IOException e) {
            throw new RuntimeException("[" + csvFilePath + "] 文件解析失败。", e);
        }
    }

    /**
     * 判断文件是否存在
     *
     * @param filePath
     * @return
     */
    public static boolean isFileExist(String filePath) {
        if (StrUtil.isEmpty(filePath)) {
            return false;
        }

        File file = new File(filePath);
        return file.exists() && !file.isDirectory();
    }

    /**
     * 分段读取大文件
     *
     * @param path         文件路径
     * @param readFileBean 分段每次读取的bean 初始值需要设置每次读取的行数
     * @param <T>          读取结果类型bean
     * @return
     */
    public <T> ReadFileBean<T> readBigFile(String path, ReadFileBean<T> readFileBean, Class<T> readCls) throws FileNotFoundException {
        File file = new File(path);
        if (!file.exists() || file.isDirectory()) {
            throw new FileNotFoundException("file:" + path + " is not found.");
        }
        try (RandomAccessFile randomAccessFile = new RandomAccessFile(file, "r")) {
            if (readFileBean.getPointer() < 0) {
                readFileBean.setPointer(0);
            }
            randomAccessFile.seek(readFileBean.getPointer());
            readFileBean.setFileLength(randomAccessFile.length());
            List<String> datas = new ArrayList<>();
            int row = 1;
            String line;
            while ((line = randomAccessFile.readLine()) != null && row <= readFileBean.getReadRowNum()) {
                row++;
                readFileBean.setPointer(randomAccessFile.getFilePointer());
                datas.add(readFileBean.lineScavenge(charsetDecode(line, readFileBean.getReadCharset())));
            }

            if (StrUtil.isEmpty(datas)) {
                //无数据
                return readFileBean.setDatas(List.of());
            }
            if (path.contains(TYPE_CSV)) {
                //csv文件处理
                com.yexuejc.base.pojo.CsvToBean csvToBean = getCsvToBean(readCls);
                readFileBean.setHeader(csvToBean.getHeader());
                if (csvToBean.isHasHeader()) {
                    //文件存在header,设置header优先,没设置使用文件的
                    if (StrUtil.isNotEmpty(csvToBean.getHeader())) {
                        //替换header
                        datas.remove(0);
                        datas.add(0, csvToBean.getHeader());
                    } else {
                        readFileBean.setHeader(datas.get(0));
                    }
                } else {
                    //文件不存在header，使用设置的
                    datas.add(0, csvToBean.getHeader());
                }

                List<T> dataList = readCsv(String.join("\n", datas), readCls, csvToBean.getDelimiter());
                readFileBean.setDatas(dataList);
            }
        } catch (FileNotFoundException e) {
            logger.severe("file exists." + e.getMessage());
        } catch (IOException e) {
            logger.severe("read file error." + e.getMessage());
        }
        return readFileBean;
    }

    /**
     * 获取csv的header,使用注解{@link CsvToBean}
     *
     * @param cls
     * @param <T>
     * @return
     */
    public static <T> com.yexuejc.base.pojo.CsvToBean getCsvToBean(Class<T> cls) {
        CsvToBean annotation = cls.getAnnotation(CsvToBean.class);
        Assert.notNull(annotation, cls.toString() + "类上需要添加注解@CsvToBean，并指定header。");
        com.yexuejc.base.pojo.CsvToBean csvToBean = new com.yexuejc.base.pojo.CsvToBean(
                annotation.header(), annotation.delimiter(), annotation.hasHeader());
        return csvToBean;
    }

    /**
     * 把字符串data按照指定编码解码
     *
     * @param data    解码字符串
     * @param charset 字符编码
     * @return
     */
    public static String charsetDecode(String data, Charset charset) {
        char[] chars = data.toCharArray();
        byte[] result = new byte[chars.length];
        for (int i = 0; i < chars.length; i++) {
            result[i] = (byte) chars[i];
        }
        return new String(result, charset);
    }
}
